#lang racket/base
;; crossfire: distributed brute force infrastructure
;;
;; Copyright (C) 2020 haskal
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU Affero General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU Affero General Public License for more details.
;;
;; You should have received a copy of the GNU Affero General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>.

(require (only-in data/integer-set make-integer-set integer-set-contents integer-set?
                  [union integer-set-union] [count integer-set-count]
                  [intersect integer-set-intersect] [subtract integer-set-subtract])
         racket/list racket/match racket/vector)

(provide char->integer-set string->integer-set range->integer-set
         ;; re-export renamed integer-set accessors
         make-integer-set integer-set-contents integer-set-count integer-set-union
         integer-set-intersect integer-set-subtract integer-set?
         pattern-count pos->pattern-pos resolve-pattern-pos
         pattern-range-take)

;; pattern processing
;; NOTE: data/integer-set WFS intervals are INCLUSIVE

;; IsetPos is a (cons inum ival)
;; inum - the interval number within the iset
;; ival - the actual input value

;; converts a position of [0, integer-set-count) to 2 values
;; - iset interval number
;; - actual value within the interval
;; this assists C code in determing where to start and end
;; (this could be moved to pure C i guess)
(define (pos->integer-set-pos iset pos)
  (define (helper wfs pos)
    (match wfs
      [(cons (cons fs fe) r)
       (define delta (- (add1 fe) fs))
       (if (< pos delta)
         (cons 0 (+ fs pos))
         (match-let ([(cons inum val) (helper r (- pos delta))])
           (cons (add1 inum) val)))]
      [_ (error "iset-pos out of range" iset pos)]))

  (helper (integer-set-contents iset) pos))

(define (char->integer-set ch)
  (define x (char->integer ch))
  (make-integer-set `((,x . ,x))))

(define (string->integer-set str)
  (for/fold ([iset (make-integer-set '())]) ([ch (in-string str)])
    (integer-set-union iset (char->integer-set ch))))

;; INCLUSIVE
(define (range->integer-set a b)
  (make-integer-set `((,a . ,b))))

;; resolved pattern pos: vector of IsetPos (obtained with pos->integer-set-pos)
;; unresolved pattern pos: vector of Nat [0, count of iset)
;; resolved pattern pos cannot be manipulated, unresolved can

;; the total input space size
(define (pattern-count pattern)
  (for/fold ([sum 1]) ([p (in-vector pattern)])
    (* sum (integer-set-count p))))

;; takes an integer in the range [0, pattern-count) and produces and unresolved pattern pos
;; this allows using integer sets to reference parts of patterns
;; produces unresolved pattern pos
(define (pos->pattern-pos pattern pos)
  (define rev-counts (reverse (for/list ([x (in-vector pattern)]) (integer-set-count x))))

  ;; bases in reverse order
  (define (helper pos bases)
    (cond
      [(zero? pos) '()]
      [(empty? bases) (error "overflowed pattern pos")]
      [else
        (define-values [q r] (quotient/remainder pos (first bases)))
        (cons r (helper q (rest bases)))]))

  (define lower-pos (list->vector (reverse (helper pos rev-counts))))
  (vector-append (make-vector (- (vector-length pattern) (vector-length lower-pos)) 0) lower-pos))

;; unresolved -> resolved
(define (resolve-pattern-pos pattern pp)
  (for/vector ([iset (in-vector pattern)] [pos (in-vector pp)])
    (pos->integer-set-pos iset pos)))

;; utils for pattern ranges [0, pattern-count) mapped to patterns with the above functions
;; a pattern range is represented by an integer-set, the wfs item boundaries would get converted
;; into resolved pattern pos, etc

;; takes up to n elements from the pattern range
;; returns the subrange and pr minus subrange
(define (pattern-range-take pr n)
  (cond
    [(>= n (integer-set-count pr)) (values pr (make-integer-set '()))]
    [else
     (match-define (cons _ bound) (pos->integer-set-pos pr (sub1 n)))
     (define subrange (integer-set-intersect (make-integer-set `((0 . ,bound))) pr))
     (define new-pr (integer-set-subtract pr subrange))
     (values subrange new-pr)]))
