#lang racket/base
;; crossfire: distributed brute force infrastructure
;;
;; Copyright (C) 2020 haskal
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU Affero General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU Affero General Public License for more details.
;;
;; You should have received a copy of the GNU Affero General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>.

(require ffi/unsafe racket/bool)
(provide get-ffi-obj/static static-ffi-available? static-ffi-arch
         ffi-lib/runtime get-ffi-obj/runtime)

;; this module provides utility functions for interacting with static ffi tables provided by an
;; embedding runtime, and for deferring either to the regular dynamic linking ffi or the static ffi
;; as the current runtime allows
;;
;; a static ffi table is a C array of structs of string and uintptr_t that define symbol exports
;; from an embedding runtime that is statically compiled (and therefore has no other usual dynamic
;; symbol table)
;;
;; applications compiled with static glibc libdl can actually still link in dynamic libraries, but
;; only if the same glibc version is available as a dynamic library during runtime. to avoid making
;; any runtime assumptions, i have created static ffi infrastructure that allows ffi calls into
;; statically compiled code exported by the embedding runtime

;; the embedding runtime provides a table in this format
(define _fun-opaque (_cpointer 'static-ffi:fun-opaque))
(define-cstruct _static-table ([name _string] [fun _fun-opaque]))

;; checks if the current runtime provides a static ffi table
(define (static-ffi-available?)
  (with-handlers ([exn:fail? (lambda (ex) #f)])
    (dynamic-require ''#%static-ffi 'table)
    #t))

;; try to load a library dynamically, but if dynamic linking is not available return #f
;; the result can be used in get-ffi-obj/runtime
(define (ffi-lib/runtime path [version #f])
  (define fail (if (static-ffi-available?) (lambda () #f) #f))
  (ffi-lib path version #:fail fail))

;; try to load an ffi object dynamically, but if lib is #f and static ffi is available, retrieve
;; the object from the static ffi table
(define (get-ffi-obj/runtime name lib type)
  (when (and (false? lib) (not (static-ffi-available?)))
    (error "no lib provided and static ffi not available"))
  (if (false? lib)
    (get-ffi-obj/static name type)
    (get-ffi-obj name lib type)))

;; retrieves the CROSSFIRE_ARCH the embedding runtime was compiled with
(define (static-ffi-arch)
  (dynamic-require ''#%static-ffi 'arch))

;; retrieve an object of a given name and type from the static ffi table, if it exists
;; otherwise raise error
(define (get-ffi-obj/static name type)
  ;; magic primitives
  (define table-addr (dynamic-require ''#%static-ffi 'table))
  (define table-size (dynamic-require ''#%static-ffi 'table-size))

  (define table (cast table-addr _uint64 _static-table-pointer))

  (define ent
    (for/fold ([res #f]) ([i (in-range table-size)])
      (define ent (ptr-ref table _static-table i))
      (if (string=? (static-table-name ent) name)
        (static-table-fun ent)
        res)))

  (when (false? ent)
    (error "static ffi entry not found" name))

  (cast ent _fun-opaque type))

;; misc ffi calls not provided by racket (because ????)
;; racket should have fsync like cmon
(module+ misc-calls
  (require ffi/unsafe/port ffi/vector racket/future racket/list racket/match racket/string)

  (provide port-fsync count-cpus current-seconds-monotonic get-hostname)

  ;; XXX : platform-specific behavior

  (define (get-libc/init)
    (let ([libc (void)])
      (when (void? libc)
        (set! libc (ffi-lib/runtime "libc" '("6" "7"))))
      libc))

  (define port-fsync/unix
    ;; lazy time
    (let ([fsync-call #f])
      (lambda (port)
        (define fd (unsafe-port->file-descriptor port))
        (when (false? fd) (error "invalid port provided"))
        (when (false? fsync-call)
          (set! fsync-call
            (get-ffi-obj/runtime "fsync" (get-libc/init)
              (_fun #:save-errno 'posix _int -> (res : _int)
                    -> (unless (zero? res)
                         (raise (exn:fail:filesystem:errno (format "fsync: errno ~a" (saved-errno))
                                                           (current-continuation-marks)
                                                           (cons (saved-errno) 'posix))))))))
        (fsync-call fd))))

  (define (port-fsync port)
    (flush-output port)
    (match (system-type 'os)
      ['unix (port-fsync/unix port)]
      [x (error "don't know how to fsync on" x)]))

  (define get-hostname/unix
    (let ([gethostname-call #f])
      (lambda ()
        (when (false? gethostname-call)
          (set! gethostname-call
            (get-ffi-obj/runtime "gethostname" (get-libc/init)
              (_fun #:save-errno 'posix (name : (_bytes o 256)) (size : _int = 256)
                    -> (res : _int)
                    -> (if (zero? res)
                         (cast name _bytes _string/utf-8)
                         (raise (exn:fail (format "gethostname: errno ~a" (saved-errno))
                                          (current-continuation-marks))))))))
        (gethostname-call))))

  (define (get-hostname)
    (match (system-type 'os)
      ['unix (get-hostname/unix)]
      [x (error "don't know how to get-hostname on" x)]))

  ;; this provides an actual count of the number of CPUs on the system, even without
  ;; --enable-futures by hooking into the underlying rktio call that gets skipped when the current
  ;; VM is configured without --enable-futures
  ;; XXX : on chez scheme, futures are always enabled (?) therefore we just use the futures call
  ;; directly. however, this hasn't been tested on ARM yet
  (define count-cpus/bc
    (let ([num-cpus #f])
      (lambda ()
        (when (false? num-cpus)
          (define self (if (static-ffi-available?) #f (ffi-lib #f)))
          ;; this should be big enough...
          (define rktio-fake (malloc 512 'atomic))
          (define rktio-init (get-ffi-obj/runtime "rktio_init_cpu" self
                             (_fun _pointer -> _void)))
          (define rktio-call (get-ffi-obj/runtime "rktio_processor_count" self
                             (_fun _pointer -> _int)))
          (rktio-init rktio-fake)
          (set! num-cpus (rktio-call rktio-fake)))
        num-cpus)))

  (define (count-cpus)
    (match (system-type 'vm)
      ['racket (count-cpus/bc)]
      ['chez-scheme (processor-count)]
      [x (error "don't know how to count-cpus on vm" x)]))


  ;; time helpers (because time is a bigge heck)
  ;; monotonic time can be different than wall clock time
  ;; for our purposes, tasks have two measures of time associated with them:
  ;;  - wall clock instant representing when the task was started
  ;;  - duration of the task (as a difference of monotonic times)
  ;; it's important never to mix up wall clock and monotonic measurements, and never to take a
  ;; difference between two wall clock times

  (define current-seconds-monotonic/unix
    (let ([clock-gettime-call #f]
          [call-param #f])
      ;; XXX : assumes time_t is long, which idk is probably a fair assumption but it could bork
      ;; things, for example on big-endian systems with 64 bit time_t and 32-bit long
      ;;; (these might not actually exist)
      (define-cstruct _timespec ([tv-sec _long] [tv-nsec _long]))
      ;; don't allocate a new timespec each call, just reuse the same one
      ;; if a thread gets preempted in this call by another thread also in this call it won't cause
      ;; any harmful effects
      (define singleton-timespec (make-timespec 0 0))
      (define *CLOCK-MONOTONIC* 1)
      (define *CLOCK-MONOTONIC-RAW* 4)
      (lambda ()
        (when (false? clock-gettime-call)
          (set! clock-gettime-call
            (get-ffi-obj/runtime "clock_gettime" (get-libc/init)
              (_fun #:save-errno 'posix _int (out : _timespec-pointer = singleton-timespec)
                    -> (res : _int)
                    -> (if (zero? res)
                         (timespec-tv-sec out)
                         (raise (exn:fail (format "clock-gettime: errno ~a" (saved-errno))
                                          (current-continuation-marks)))))))
          ;; attempt with linux-specific parameter
          ;; if it fails, fall back to POSIX
          (with-handlers ([exn:fail? (lambda (_) (set! call-param *CLOCK-MONOTONIC*))])
            (clock-gettime-call *CLOCK-MONOTONIC-RAW*)
            (set! call-param *CLOCK-MONOTONIC-RAW*)))
        (clock-gettime-call call-param))))

  (define (current-seconds-monotonic)
    (match (system-type 'os)
      ['unix (current-seconds-monotonic/unix)]
      [x (error "don't know how to current-seconds-monotonic on" x)])))
